<?php

App::uses('CakeEventListener', 'Event');
App::uses('DiscountLib', 'EvDiscount.Lib');

class EvBasketCheckForDiscountListener implements CakeEventListener {

	public function implementedEvents() {
		return array(
			'EvDiscount.Component.DiscountCodes.codeAdded' => 'addDiscount',
			'EvBasket.Component.Basket.itemAdd' => 'rebuildDiscount',
			'EvBasket.Component.Basket.itemUpdated' => 'rebuildDiscount',
			'EvBasket.Component.Basket.itemDeleted' => 'rebuildDiscount',
			'EvLoyaltyPoints.Component.Points.usePoints' => 'addLoyaltyDiscount',
			'EvBasket.Component.Basket.totalRowUpdated' => 'basketTotalRowUpdated',
		);
	}

	/**
	 * queue up the confirmation emails
	 *
	 * @param   CakeEvent
	 */
	public function addDiscount(CakeEvent $Event, $data = null) {
		$dataCode = $Event->data['data'];
		// Event subject will be the EvDiscount.DiscountCodesComponent
		return $this->rebuildDiscount($Event, $dataCode);
	}

	/**
	 * Update discounts if a row total changes. This function filters out the discount row change to avoid
	 */
	public function basketTotalRowUpdated(CakeEvent $Event) {
		$updateData = $Event->data['data'];

		if ($updateData['name'] == Configure::read('EvBasket.labels.delivery')) {
			return $this->rebuildDiscount($Event);
		} else {
			return true;
		}
	}

	/**
	 * rebuild discount due to basket amends
	 *
	 * @param   CakeEvent
	 */
	public function rebuildDiscount(CakeEvent $Event, $dataCode = null) {
		if (! CakePlugin::loaded('EvDiscount')) {
			return false;
		}

		$this->_addFreeItemsFromDiscounts($Event);
		$this->_addFreeItemsFromPromotions($Event);

		$EventSubject = $Event->subject();

		$basket = $EventSubject->_controller->BasketManager->getBasket();

		$priceModifier = $this->_setupPriceModifier($basket);

		$this->_populatePriceModifierBasket($priceModifier, $basket);

		$this->_populatePriceModifierDiscounts($priceModifier, $EventSubject, $dataCode);
		$this->_populatePriceModifierPromotions($priceModifier, $EventSubject);

		// Make the price modifier apply the discounts and get the results back
		$priceModifier->applyDiscounts();
		$discountedBasket = $priceModifier->basket()->toArray();

		//Default discount position after subtotal. Discount will always contain what is to betaken off of the subtotal.
		//Tax amounts always contain the discounted value so display this after discount.
		$discountSequence = 15;

		// Loop through the discount breakdown and add up those for vouchers and those for discounts separately
		// Add a basket data record to handle the discount row
		$voucherPreTaxTotal = $voucherIncTaxTotal = $discountPreTaxTotal = $discountIncTaxTotal = 0;
		foreach ($discountedBasket['discountBreakdown'] as $discountCode => $discountAmount) {
			// Voucher codes get prefixed with V_ at the top of this function when adding them
			if (strpos($discountCode, "V_") === 0) {
				$EventSubject->_controller->BasketManager->setBasketData('voucher_code', str_replace("V_", "", $discountCode), true);
				$voucherPreTaxTotal += $discountAmount['amount'];
				$voucherIncTaxTotal += $discountAmount['amount'] + $discountAmount['tax'];
			} elseif (strpos($discountCode, "promotion_") === 0) {
				$EventSubject->_controller->BasketManager->setBasketData('promotion_id', str_replace("promotion_", "", $discountCode), true);
				$discountPreTaxTotal += $discountAmount['amount'];
				$discountIncTaxTotal += $discountAmount['amount'] + $discountAmount['tax'];
			} else {
				$EventSubject->_controller->BasketManager->setBasketData('discount_code', $discountCode, true);
				$discountPreTaxTotal += $discountAmount['amount'];
				$discountIncTaxTotal += $discountAmount['amount'] + $discountAmount['tax'];
			}
		}

		// Add Voucher discount total
		if ($voucherPreTaxTotal > 0) {
			//Don't update the tax as that is dealt with when rebuilding totals
			$EventSubject->_controller->BasketManager->manageTotalRow(
				Configure::read('EvBasket.labels.voucher'),
				-$voucherPreTaxTotal,
				$discountSequence,
				false,
				-$voucherPreTaxTotal,
				-$voucherIncTaxTotal
			);
		} else {
			// No discount is applicable to the basket remove the row
			$EventSubject->_controller->BasketManager->deleteTotalRow(Configure::read('EvBasket.labels.voucher'));
		}

		// Add Discount discount total
		if ($discountPreTaxTotal > 0) {
			//Don't update the tax as that is dealt with when rebuilding totals
			$EventSubject->_controller->BasketManager->manageTotalRow(
				Configure::read('EvBasket.labels.discount'),
				-$discountPreTaxTotal,
				$discountSequence,
				false,
				-$discountPreTaxTotal,
				-$discountIncTaxTotal
			);
		} else {
			// No discount is applicable to the basket remove the row
			$EventSubject->_controller->BasketManager->deleteTotalRow(Configure::read('EvBasket.labels.discount'));
		}

		// Now we have added the discount row we can rebuild the totals
		$Event->subject()->_controller->BasketManager->rebuildTotals();

		return true;
	}

	public function addLoyaltyDiscount(CakeEvent $Event) {
		if (! CakePlugin::loaded('EvLoyaltyPoints')) {
			return false;
		}

		$EventSubject = $Event->subject();

		if (! is_object($EventSubject->_controller->BasketManager)) {
			$EventSubject->_controller->BasketManager = $EventSubject->_controller->loadComponent('EvBasket.BasketManager');
		}

		$total = $EventSubject->_controller->BasketManager->findTotalRow(
			Configure::read('EvBasket.labels.subtotal')
		);

		App::uses('PointLib', 'EvLoyaltyPoints.Lib');

		if ($Event->data['pointsToUse'] <= 0) {
			$discount = 0;
		} else {
			$discount = PointLib::calculate($total['amount'], 'EvBasket', 0, $Event->data['pointsToUse']);
		}

		$EventSubject->_controller->BasketManager->manageTotalRow(
			Configure::read('EvBasket.labels.discount'),
			-$discount,
			15
		);

		// Add a basket data record to handle the loyalty disount row
		$EventSubject->_controller->BasketManager->setBasketData('loyalty_points_used', $Event->data['pointsToUse'], true);

		$Event->subject()->_controller->BasketManager->rebuildTotals();

		return true;
	}

/**
 * Sets up the price modifier configuration and returns a setup version
 * @param  array  			$basket Basket array
 * @return PriceModifier    Initialised price modifier object
 */
	protected function _setupPriceModifier($basket) {
		$priceModifier = new Evoluted\PriceModifier\PriceModifier(
			new Evoluted\PriceModifier\Basket\Basket(new Evoluted\PriceModifier\Storage\Runtime, $basket['Basket']['id'])
		);

		require_once (CakePlugin::path('EvDiscount') . 'Lib' . DS . 'DiscountRestrictions.php');
		$priceModifier->setDiscountRestrictionHandler('Evoluted\EvDiscount\DiscountRestrictions\DiscountRestrictions');

		// In the case of our shop we want vouchers to actually go past the subtotal amount so they apply to delivery as well.
		require_once (CakePlugin::path('EvDiscount') . 'Lib' . DS . 'VoucherDiscountModifier.php');
		$priceModifier->addDiscountModifier('voucher', 'Evoluted\PriceModifier\DiscountModifiers\VoucherDiscountModifier');

		require_once (CakePlugin::path('EvDiscount') . 'Lib' . DS . 'CheapestFreeDiscountModifier.php');
		$priceModifier->addDiscountModifier('Buy X get cheapest free', 'Evoluted\PriceModifier\DiscountModifiers\CheapestFreeDiscountModifier');

		require_once (CakePlugin::path('EvDiscount') . 'Lib' . DS . 'CheapestHalfPriceDiscountModifier.php');
		$priceModifier->addDiscountModifier('Buy X get cheapest half price', 'Evoluted\PriceModifier\DiscountModifiers\CheapestHalfPriceDiscountModifier');

		require_once (CakePlugin::path('EvDiscount') . 'Lib' . DS . 'XForYDiscountModifier.php');
		$priceModifier->addDiscountModifier('X for the price of Y', 'Evoluted\PriceModifier\DiscountModifiers\XForYDiscountModifier');

		require_once (CakePlugin::path('EvDiscount') . 'Lib' . DS . 'FreeDeliveryDiscountModifier.php');
		$priceModifier->addDiscountModifier('Free delivery', 'Evoluted\PriceModifier\DiscountModifiers\FreeDeliveryDiscountModifier');

		require_once (CakePlugin::path('EvDiscount') . 'Lib' . DS . 'FreeGiftDiscountModifier.php');
		$priceModifier->addDiscountModifier('Free gift', 'Evoluted\PriceModifier\DiscountModifiers\FreeGiftDiscountModifier');

		// Set whether the basket is an inc or exvat basket
		$priceModifier->applyDiscountsAfterTax = Configure::read('SiteSetting.ev_basket.discounts_include_tax');

		return $priceModifier;
	}

/**
 * Populates the price modifier basket with basket items. Modifies the price modifier directly
 * @param  PriceModifier 	&$priceModifier Price modifier object
 * @param  array 			$basket         Basket Array
 * @return null
 */
	protected function _populatePriceModifierBasket(&$priceModifier, $basket) {
		// Add each basket item to the price modifier
		foreach ($basket['BasketItem'] as $basketItem) {
			$priceModifier->basket()->insert([
				'id' => $basketItem['id'],
				'unitPrice' => $basketItem['unit_price'],
				'quantity' => $basketItem['quantity'],
				'taxRate' => $basketItem['tax_rate'],
				'model' => $basketItem['model'],
				'model_id' => $basketItem['model_id'],
				'Variant' => $basketItem['Variant']
			]);
		}
	}

/**
 * Adds the dicount rules from the curent dicount codes
 * @param  PriceModifier 	&$priceModifier Price modifier object
 * @param  array 			$basket         Basket Array
 * @return null
 */
	protected function _populatePriceModifierDiscounts(&$priceModifier, &$EventSubject, $dataCode) {
		// The cookie isn't used any more but there can still be multiple discount codes added to the basket in the
		// database potentially. Current functionality only allows one discount code at a time though.
		if (! is_object($EventSubject->_controller->BasketManager)) {
			$EventSubject->_controller->BasketManager = $EventSubject->_controller->loadComponent('EvBasket.BasketManager');
		}

		// Get the basket data to find the discount codes
		$codes = array();
		if ($dataCode != null) {
			$codes[$dataCode['DiscountCode']['code']] = $dataCode;
		}

		// Add additional discount codes from basket data
		$basketData = $EventSubject->_controller->BasketManager->getBasketData();
		if (!empty($basketData)) {
			foreach ($basketData as $key => $value) {
				if ($value['name'] == 'discount_code' || $value['name'] == 'voucher_code') {
					$codes[$value['data']] = $value;
				}
			}
		}

		$deliveryTotal = $EventSubject->_controller->BasketManager->findTotalRow(Configure::read('EvBasket.labels.delivery'));

		// If the basket data returns some codes or one is passed through (or both) then
		if (! empty($codes) && is_array($codes)) {
			// Checks with the discount module to see what codes are active and not run out etc.
			// Only returns codes that are still valid
			$DiscountCode = EvClassRegistry::init('EvDiscount.DiscountCode');
			$validCodes = $DiscountCode->validateCodes($codes);

			if (! empty($validCodes)) {
				// If there were any valid discount modifiers add to the discount rules for the price modifier.
				foreach ($validCodes as $code => $codeData) {
					// Add discount rule
					if ($codeData['DiscountCode']['is_voucher']) {
						$discountType = 'voucher';
					} else {
						$discountType = $codeData['DiscountType']['name'];
					}

					$priceModifier->addDiscountRule([
						'id' => $codeData['DiscountCode']['is_voucher'] ? ("V_" . $code) : $code,
						'discountType' => $discountType,
						// Used for percentage discounts
						'percent' => $codeData['DiscountCode']['amount'],
						// Used for fixed discounts (ignored by percentage so fine to always pass)
						'amount' => $codeData['DiscountCode']['amount'],
						// Percentage discounts are applied per item and fixed discounts are applied at the end
						'applyToItems' => $codeData['DiscountType']['apply_to_items'],
						// Apply discounts before or after applying tax
						'applyDiscountsAfterTax' => Configure::read('SiteSetting.ev_basket.discounts_include_tax'),
						// Additional information passed through for the price modifiers restrictions
						'min_order' => $codeData['DiscountCode']['min_order'],
						'code_restrictions' => $codeData['CodeRestriction'],
						'extra_data' => !empty($codeData['DiscountCode']['extra_data']) ? json_decode($codeData['DiscountCode']['extra_data'], true) : null,

						'delivery' => !empty($deliveryTotal) ? $deliveryTotal['amount'] : 0,
						'deliveryTax' => !empty($deliveryTotal) ? $deliveryTotal['tax'] : 0,
					]);
				}
			}
		}
	}

/**
 * Adds the discount rules from any active promotions
 * @param  PriceModifier 	&$priceModifier Price modifier object
 * @param  array 			$basket         Basket Array
 * @return null
 */
	protected function _populatePriceModifierPromotions(&$priceModifier, &$EventSubject) {
		if (!CakePlugin::loaded('EvPromotion')) {
			// Not using promotions
			return false;
		}

		if (! is_object($EventSubject->_controller->BasketManager)) {
			$EventSubject->_controller->BasketManager = $EventSubject->_controller->loadComponent('EvBasket.BasketManager');
		}

		$Promotion = EvClassRegistry::init('EvPromotion.Promotion');
		$activePromotions = $Promotion->findAllActive();

		$deliveryTotal = $EventSubject->_controller->BasketManager->findTotalRow(Configure::read('EvBasket.labels.delivery'));

		// If there are any active promotions add the price modifiers
		if (!empty($activePromotions)) {
				// If there were any valid discount modifiers add to the discount rules for the price modifier.
			foreach ($activePromotions as $promotion) {
				$priceModifier->addDiscountRule([
					'id' => 'promotion_' . $promotion['Promotion']['id'],
					'discountType' => $promotion['DiscountType']['name'],
					// Used for percentage discounts
					'percent' => $promotion['Promotion']['amount'],
					// Used for fixed discounts (ignored by percentage so fine to always pass)
					'amount' => $promotion['Promotion']['amount'],
					// Percentage discounts are applied per item and fixed discounts are applied at the end
					'applyToItems' => $promotion['DiscountType']['apply_to_items'],
					// Apply discounts before or after applying tax
					'applyDiscountsAfterTax' => Configure::read('SiteSetting.ev_basket.discounts_include_tax'),
					// Additional information passed through for the price modifiers restrictions
					'min_order' => $promotion['Promotion']['min_order'],
					'code_restrictions' => $promotion['CodeRestriction'],

					'extra_data' => !empty($promotion['Promotion']['extra_data']) ? json_decode($promotion['Promotion']['extra_data'], true) : null,

					'delivery' => !empty($deliveryTotal) ? $deliveryTotal['amount'] : 0,
					'deliveryTax' => !empty($deliveryTotal) ? $deliveryTotal['tax'] : 0,
				]);
			}
		}
	}

/**
 * Adds some free items to the basket if other qualifying items are present
 */
	protected function _addFreeItemsFromDiscounts($Event) {
		$EventSubject = $Event->subject();
		// The cookie isn't used any more but there can still be multiple discount codes added to the basket in the
		// database potentially. Current functionality only allows one discount code at a time though.
		if (! is_object($EventSubject->_controller->BasketManager)) {
			$EventSubject->_controller->BasketManager = $EventSubject->_controller->loadComponent('EvBasket.BasketManager');
		}

		// Get discount codes from basket data
		$codes = array();
		$basketData = $EventSubject->_controller->BasketManager->getBasketData();
		if (!empty($basketData)) {
			foreach ($basketData as $key => $value) {
				if ($value['name'] == 'discount_code' || $value['name'] == 'voucher_code') {
					$codes[$value['data']] = $value;
				}
			}
		}

		// If the basket data returns some codes or one is passed through (or both) then
		if (! empty($codes) && is_array($codes)) {
			// Checks with the discount module to see what codes are active and not run out etc.
			// Only returns codes that are still valid
			$DiscountCode = EvClassRegistry::init('EvDiscount.DiscountCode');
			$validCodes = $DiscountCode->validateCodes($codes);

			if (!empty($validCodes)) {

				$basket = $EventSubject->_controller->BasketManager->getBasket();

				// If there were any valid codes...
				foreach ($validCodes as $code => $codeData) {

					if ($codeData['DiscountType']['name'] == 'Free gift') {
						$this->_addFreeProducts( $EventSubject->_controller->BasketManager, $codeData['DiscountData']['free_gift_product_id'][0], $codeData['CodeRestriction'], $codeData['DiscountCode']['amount'] );
					}
				}
			}
		}
	}

/**
 * Adds some free items to the basket if other qualifying items are present using the active promotions
 */
	protected function _addFreeItemsFromPromotions($Event) {
		if (!CakePlugin::loaded('EvPromotion')) {
			// Not using promotions
			return false;
		}

		$EventSubject = $Event->subject();
		if (! is_object($EventSubject->_controller->BasketManager)) {
			$EventSubject->_controller->BasketManager = $EventSubject->_controller->loadComponent('EvBasket.BasketManager');
		}

		// Get discount codes from basket data
		$Promotion = EvClassRegistry::init('EvPromotion.Promotion');
		$activePromotions = $Promotion->findAllActive();

		// If there are active promotions
		if (!empty($activePromotions)) {

			// If there were any...
			foreach ($activePromotions as $promoData) {
				if ($promoData['DiscountType']['name'] == 'Free gift') {
					$this->_addFreeProducts( $EventSubject->_controller->BasketManager, $promoData['DiscountData']['free_gift_product_id'][0], $promoData['CodeRestriction'], $promoData['Promotion']['amount'] );
				}
			}
		}
	}

	protected function _addFreeProducts($BasketManager, $freeProductId, $restrictions, $amountToBuyPerFreeProduct) {
		//
		$currentFreebies = 0;
		$totalCount = 0;

		$basket = $BasketManager->getBasket();

		// Check we have the required items in the basket.
		foreach ($basket['BasketItem'] as $basketItem) {

			// Get the product id (depends on the type of basket item)
			if ($basketItem['model'] == 'EvShop.Variant') {
				$productId = $basketItem['Variant']['Product']['id'];
			} else {
				// Not interested in this item
				continue;
			}

			$checkDiscountRestriction = new \CakeEvent('EvBasket.Component.BasketItem.CheckDiscountRestrictions', $basketItem, array(
				'discount' => [
					'code_restrictions' => $restrictions,
				],
				'basket_item' => $basketItem,
			));
			\CakeEventManager::instance()->dispatch($checkDiscountRestriction);
			// If false is returned discount isn't available for this item
			$result = $checkDiscountRestriction->result;

			if ($productId == $freeProductId) {
				$currentFreebies += $basketItem['quantity'];
			} elseif (!empty($result['discountValue'])) {
				$totalCount += $basketItem['quantity'];
			}

		}

		$requiredFreebies = floor($totalCount / $amountToBuyPerFreeProduct) - $currentFreebies;
		if ($requiredFreebies > 0) {
			$this->_addProductToBasket($BasketManager, $freeProductId, $requiredFreebies);
		}
	}

	/**
	 * Adds a product to the basket. Will add the first available variant.
	 */
	protected function _addProductToBasket($BasketManager, $productId, $quantity) {
		//
		$useInventory = false;
		$productQuery = [
			'conditions' => [
				'Product.id' => $productId
			],
			'contain' => [
				'Variant'
			]
		];

		if (CakePlugin::loaded('EvInventory')) {
			// Only contain inventory if the plugin is loaded
			$useInventory = true;
			$productQuery['contain']['Variant'] = 'Inventory';
		}

		$Product = EvClassRegistry::init('EvShop.Product');
		$product = $Product->find('first', $productQuery);

		$totalRemaining = $quantity;
		foreach ($product['Variant'] as $variant) {
			if (!$useInventory || $variant['Inventory']['stock'] >= $quantity) {
				// Assign a block of this variant and stop there
				$this->_addVariantToBasket($BasketManager, $variant['id'], $totalRemaining);
				$totalRemaining = 0;
			} elseif ($variant['Inventory']['stock'] > 0) {
				// Assign as many as we can of this variant and move on to the next
				$this->_addVariantToBasket($BasketManager, $variant['id'], $variant['Inventory']['stock']);
				$totalRemaining -= $variant['Inventory']['stock'];
			}

			if ($totalRemaining <= 0) {
				// We've added all we need to
				break;
			}
		}
	}

	protected function _addVariantToBasket($BasketManager, $variantId, $quantity) {
		// setup the data
		$item = array(
			'model' => 'EvShop.Variant',
			'model_id' => $variantId
		);

		$Model = EvClassRegistry::init('EvShop.Variant');
		$unitPrice = $Model->getUnitPrice($variantId);

		// if there's a method to get the taxRate, do so
		$taxRate = 0;
		if (method_exists($Model, 'getTaxRate')) {
			$taxRate = $Model->getTaxRate($item['model_id']);
		}

		if ($unitPrice !== false) {
			// add to the basket but dont dispatch the event or we'll end up going around in circles.
			$BasketManager->addItem($item, $unitPrice, $quantity, $taxRate, false);
		}
	}
}
